#include "jengine.h"
#include "util.h"
#include "sha1.h"
#include <string>
#include <iostream>
#include <fstream>

#include <sys/types.h>
#include <sys/stat.h>

using namespace std;

static const char *cache_dir="./yacts.cache";

// compute the hash of verb (given as string) and its input and the 
// corresponding cache file name
std::string cache_fname(jarray verb, jarray x, jarray y) {
  SHA1 sha1;
  if (x.isValid()) x.addhash(sha1);
  y.addhash(sha1);
  verb.addhash(sha1);
  string fname(cache_dir);
  fname+='/';
  fname+=sha1tostring(sha1);
  return fname;
};

jarray loadFromCache(jarray verb, jarray x, jarray y) {
  jengine *je=y.getEngine();
  jarray r;
  // check that cache file exists
  ifstream icachefile;
  icachefile.open(cache_fname(verb, x,y).c_str(), ios::in | ios::binary);
  if (icachefile.is_open()) { // cache file found, check arguments exactly
    jarray verbverb(je, icachefile);
    jarray xx;
    if (x.isValid()) {
      jarray xxx(je, icachefile);
      xx=xxx;
    };
    jarray yy(je,icachefile);
    if ((verb==verbverb)&&
	((!x.isValid())||(x==xx))&&(y==yy)) { // proceed with loading the result
      jarray rr(je, icachefile);
      if (rr.isValid() && icachefile.good()) r=rr; // cache hit
    };
    icachefile.close();
  };
  return r;
};

void storeInCache(jarray verb, jarray x, jarray y, jarray r) {
  ofstream ocachefile;
  ocachefile.open(cache_fname(verb, x, y).c_str(), ios::out | ios::binary);
  if (!ocachefile.is_open()) { // create directory and try again
    if (mkdir(cache_dir, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH)==0)
      ocachefile.open(cache_fname(verb, x, y).c_str(), ios::out | ios::binary);
  };
  if (!ocachefile.is_open()) return; // can't write, give up
  verb.write(ocachefile);
  if (x.isValid()) x.write(ocachefile);
  y.write(ocachefile);
  r.write(ocachefile);
  ocachefile.close();
};

jarray memo(jengine::monad um, jengine::dyad ud, jarray x, jarray y) {
  jarray dummy; // invalid array
  jarray verb=ud(y,dummy);
  jarray r=loadFromCache(verb, x, y);
  if (!r.isValid()) { // cache miss
    r=x.isValid()?ud(x,y):um(y); // compute the verb
    storeInCache(verb, x, y, r);
  };
  return r;
};

jarray memo(jengine::monad um, jengine::dyad ud, jarray y) {
  jarray x; // dummy invalid array
  return memo(um, ud, x, y);
};
